#include "connections_handler.h"


//prototype static functions
static void * __runConnectionsHandler(void *ptr);
static int socketWriteData(unsigned char *data, int length, int sock);

/**********************************************************************************/
/**********************************************************************************/

void initialiseConnectionStruct(connections_t *con){
	//
	// Function to initialise a connections structure. This must be 
	// called before the structure is used.
	//
	int i;
	pthread_mutex_init(&(con->mutex),NULL);
	for(i=0; i<MAX_NUM_CLIENTS; i++){
		con->sockfds[i] = -1;
	}	
}

/**********************************************************************************/
/**********************************************************************************/

void runConnectionsHandler(connections_t *con){
	//
	// Function launches new thread which listens for client connections
	// and maintains the connections_t structure to reflect what clients
	// are currently connected. The connections_t structure passed to this
	// function must have been initialised!
	//
	pthread_t connection_handler_thread;
	pthread_attr_t attr;

	//initialise variables
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	//start a new thread to handle client connections
	pthread_create(&connection_handler_thread, &attr, __runConnectionsHandler, con);
}

/**********************************************************************************/
/**********************************************************************************/

void sendSensorData(connections_t *con, double voltage, double current, double intT, double extT){
	double sensor_data[4] = {voltage, current, intT, extT};
	int i;
	//send the sensor data to all clients
	pthread_mutex_lock(&(con->mutex));
	for (i=0; i< MAX_NUM_CLIENTS; i++){
		if (con->sockfds[i]>0){
			socketWriteData((unsigned char *)sensor_data, 4*sizeof(double), con->sockfds[i]);
			close(con->sockfds[i]);
			con->sockfds[i] = -1;
			con->count--;
		}
	}
	pthread_mutex_unlock(&(con->mutex));
	
}

/**********************************************************************************/
/**********************************************************************************/
//static function
void * __runConnectionsHandler(void *ptr){
	//
	// Worker function that is run in a new thread (launched by runConnectionHandler).
	// Accepts client connections and stores the file descriptors in a list
	//
	connections_t *connections = (connections_t *)ptr;
	int sockfd, newsockfd, portno;
	int i; //counter

	socklen_t clilen;
	struct sockaddr_in serv_addr, cli_addr;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sockfd < 0){
		FATAL("ERROR opening socket");
	}

	//set SO_REUSEADDR so that the server can come back up immediately after a crash
	i = 1;
	if(setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&i, sizeof(i)) < 0){
		FATAL("Setting SO_REUSEADDR error");
	}

	bzero((char *) &serv_addr, sizeof(serv_addr));

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(PORT_NUM);

	if (bind(sockfd, (struct sockaddr *) &serv_addr,
			  sizeof(serv_addr)) < 0){
			  close(sockfd);
			  FATAL("ERROR on binding");
	}
	listen(sockfd,5);
	clilen = sizeof(cli_addr);

	while (1) {
		 newsockfd = accept(sockfd,
			   (struct sockaddr *) &cli_addr, &clilen);
		 if (newsockfd < 0){
			 close(sockfd);
			 FATAL("ERROR on accept");
		 }

		 //look for a free space in the connections list
		 pthread_mutex_lock(&(connections->mutex));
		 i = 0;
		 while (i < MAX_NUM_CLIENTS){
			 if (connections->sockfds[i] < 0){
				 connections->sockfds[i] = newsockfd;
				 connections->count += 1;
				 break;
			 }
			 i++;
		 }
		 pthread_mutex_unlock(&(connections->mutex));

		 if (i == MAX_NUM_CLIENTS){
			 //reject the connection
			 close(newsockfd);
		 }

	 } /* end of while */
	 close(sockfd);
	 return 0; /* we never get here */
}

/**********************************************************************************/
/**********************************************************************************/

int socketWriteData(unsigned char *data, int length, int sock){
	//
	// Attempts to write "length" bytes of "data" to the socket. Returns 0
	// on success, 1 on failure
	//
	int n;
	int total_n=0;

	do {
		// use MSG_NOSIGNAL flag to prevent OS killing the server if a
		// client disconnects during a write
		n = send(sock, data+total_n, length - total_n, MSG_NOSIGNAL);
		if (n <= 0){
			return 1;
		}
		total_n += n;

	} while (total_n < length);
	return 0;
}

/**********************************************************************************/
/**********************************************************************************/
