#include "led_control.h"
#include "logging.h"

//prototype static functions
static void * __runLEDIndicator(void *x);
static void setLED(int status);


/**********************************************************************************/
/**********************************************************************************/

void initialiseLEDStruct(led_ctrl_t *led){
	//
	// Function to initialise an led_ctrl_t structure. This must be 
	// called before the structure is used.
	//
	pthread_mutex_init(&(led->status_mutex),NULL);
	pthread_cond_init(&(led->status_cond_var), NULL);
	led->status = NONE;
}

/**********************************************************************************/
/**********************************************************************************/

#ifndef USE_GPIO
void runLEDIndicator(led_ctrl_t *led_ctrl){}
void signalNewImage(led_ctrl_t *led_ctrl){}
#else
void runLEDIndicator(led_ctrl_t *led_ctrl){
	//
	// Function launches new thread which flashes the LED each time
	// it is signalled to do so with the signalNewImage function.
	// The led_ctrl_t structure must have been initialised before it
	// is passed to this function.
	//
	pthread_t led_thread;
	pthread_attr_t attr;

	//initialise variables
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	//start a new thread to handle client connections
	pthread_create(&led_thread, &attr, __runLEDIndicator, led_ctrl);
}

/**********************************************************************************/
/**********************************************************************************/

void signalNewImage(led_ctrl_t *led_ctrl){
	//
	// Signal to the led thread that a new image has arrived. Assuming that the 
	// thread is not busy, then this will cause the led to be flashed.
	//
	DEBUG("Signalling new image has arrived");
	pthread_mutex_lock(&(led_ctrl->status_mutex));

	if (led_ctrl->status == WAITING){
		led_ctrl->status = IMAGE_RECIEVED;
		pthread_cond_signal(&(led_ctrl->status_cond_var));
	} else {
		led_ctrl->status = IMAGE_RECIEVED;
	}
	pthread_mutex_unlock(&(led_ctrl->status_mutex));
}


/**********************************************************************************/
/**********************************************************************************/

//static function
void * __runLEDIndicator(void *x){
	//
	// Worker function that is run in a new thread (launched by runLEDIndicator).
	// Flashes the LED when signalled to do so by calls to signalNewImage.
	//
	
	//initialise the gpio library
	if (gpio_init()){
		WARN("Failed to initialise the userspace gpio library.");
		pthread_exit(NULL);
	}
	
	//request the led gpio pin
	if (gpio_request(LED_GPIO_PIN, "led_gpio") < 0){
		//failed to get it - maybe it is busy? Try freeing it first
		gpio_free(LED_GPIO_PIN);
		if (gpio_request(LED_GPIO_PIN, "led_gpio") < 0){
			//ok still can't get it, give up
			WARN("Failed to acquire GPIO pin %d",LED_GPIO_PIN);
			pthread_exit(NULL);
		}
	}
	
	//set the gpio pin for output, and initialise it to off
	gpio_direction_output(LED_GPIO_PIN, GPIO_OFF);
	
	//wait for signals.
	led_ctrl_t *led_ctrl = x;
	pthread_mutex_lock(&(led_ctrl->status_mutex));
	while (led_ctrl->status != EXIT_NOW){
		if (led_ctrl->status == IMAGE_RECIEVED){
			led_ctrl->status = PROCESSING;
			pthread_mutex_unlock(&(led_ctrl->status_mutex));
			gpio_set_value(LED_GPIO_PIN, GPIO_ON);
			usleep(LED_TIME_ON*1000);
			gpio_set_value(LED_GPIO_PIN, GPIO_OFF);
			pthread_mutex_lock(&(led_ctrl->status_mutex));
		} else {
			led_ctrl->status = WAITING;
			pthread_cond_wait(&(led_ctrl->status_cond_var), &(led_ctrl->status_mutex));
		}
	}
	gpio_free(LED_GPIO_PIN);
	gpio_term();
	pthread_mutex_unlock(&(led_ctrl->status_mutex));
	pthread_exit(NULL);
}
#endif //USE_GPIO
/**********************************************************************************/
/**********************************************************************************/

