// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LOGGING_H_
#define LOGGING_H_
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>

#include"str_functions.h"

//TODO
//the logging functions are not thread safe at the moment - so just disable logfile writing for now!
#ifdef LOGFILE
#undef LOGFILE
#endif


#define MAX_LOGFILE_SIZE 2.0 //max file size in megabytes


static char __logging_timestr[50];
static char __logging_filename[50];
static int __logging_linenumber;



static inline void setLoggingTimestamp(void){
	struct timeval tv;
	gettimeofday(&tv,NULL);
	stringFromTime(__logging_timestr, 49, "%d/%m/%Y %H:%M:%S.%s",tv.tv_sec, (int)(tv.tv_usec));
}

#define SET_PREFIX_DATA setLoggingTimestamp(); strcpy(__logging_filename, __FILE__); __logging_linenumber= __LINE__;



// Macros for outputting status messages at different levels of verbosity
// Note that FATAL is always defined - presumably if something fatal happens,
// then you always want the program to exit?!
#define LVL_DEBUG 2
#define LVL_WARN 1

// if no verbosity level was specified, then we only output fatal errors
#ifndef VERBOSITY
#define VERBOSITY LVL_DEBUG
#endif

#ifdef LOGFILE
//function to move full log files - we only keep one old log file.
static inline void _rollLogFile(void){

	if(rename(LOGFILE, LOGFILE".0") < 0){
		//if the rename fails for some reason, then try just opening the
		//file in write mode - this will overwrite the old log file, but
		//this is better than having a logfile that continues to get bigger!
		FILE *fp = fopen(LOGFILE,"w");
		fclose(fp);
	}

}
#endif

static inline void _doNothing(const char *message,...){};


// define functions to actually print and/or log the messages
static inline void _logMessage(const char *message,...){
    va_list args;
    va_start(args, message);
#ifdef LOGFILE
    FILE *ofp = fopen(LOGFILE,"a");
    if (ofp != NULL){
    		fprintf(ofp, "[%s] ", __logging_timestr);
    		vfprintf(ofp, message, args);
    		fprintf(ofp, "\n");

    		if (ftell(ofp) > MAX_LOGFILE_SIZE*1024*1024){
    			fclose(ofp);
    			_rollLogFile();
    		}else{
    			fclose(ofp);
    		}
    }else{
    	error(0, 0, "Failed to open %s for writing log message.\n", LOGFILE);
    }
#else
    printf("[%s] ",__logging_timestr);
    vprintf(message,args);
    printf("\n");
#endif
    va_end(args);
}


#define LOGMESSAGE  SET_PREFIX_DATA _logMessage

// DEBUG messages
#if VERBOSITY >= LVL_DEBUG
#define DEBUG  SET_PREFIX_DATA _debugMessage

static inline void _debugMessage(const char *message,...){
    va_list args;
    va_start(args, message);
#ifdef LOGFILE
    FILE *ofp = fopen(LOGFILE,"a");
    if (ofp != NULL){
    		fprintf(ofp, "[%s]", __logging_timestr);
    		fprintf(ofp, "(%s:%d)", __logging_filename, __logging_linenumber);
    		fprintf(ofp," * Debug *: ");
    		vfprintf(ofp, message, args);
    		fprintf(ofp, "\n");

    		if (ftell(ofp) > MAX_LOGFILE_SIZE*1024*1024){
    			fclose(ofp);
    			_rollLogFile();
    		}else{
    			fclose(ofp);
    		}
    }else{
    	error(0, 0, "Failed to open %s for writing debug message.\n", LOGFILE);
    }
#else
    printf("[%s]",__logging_timestr);
    printf("(%s:%d)", __logging_filename, __logging_linenumber);
    printf(" * Debug *: ");
    vprintf(message,args);
    printf("\n");
#endif
    va_end(args);
}
#else
#define DEBUG _doNothing
#endif


// WARNING messages
#if VERBOSITY >= LVL_WARN
#define WARN  SET_PREFIX_DATA _warningMessage

static inline void _warningMessage(const char *message,...){
    va_list args;
    va_start(args, message);

#ifdef LOGFILE
    FILE *ofp = fopen(LOGFILE,"a");
    if (ofp != NULL){
    		fprintf(ofp, "[%s]", __logging_timestr);
    		fprintf(ofp, "(%s:%d)", __logging_filename, __logging_linenumber);
    		fprintf(ofp," ** Warning! **: ");
    		vfprintf(ofp, message, args);
    		fprintf(ofp, "\n");

    		if (ftell(ofp) > MAX_LOGFILE_SIZE*1024*1024){
    			fclose(ofp);
    			_rollLogFile();
    		}else{
    			fclose(ofp);
    		}
    }else{
    	error(0, 0, "Failed to open %s for writing warning message.\n", LOGFILE);
    }
#else
    printf("[%s]",__logging_timestr);
    printf("(%s:%d)", __logging_filename, __logging_linenumber);
    printf(" ** Warning! **: ");
    vprintf(message,args);
    printf("\n");
#endif
    va_end(args);
}

#else
	#define WARN _doNothing
#endif



//FATAL messages

#define FATAL SET_PREFIX_DATA _errorMessage

static inline void _errorMessage(const char *message,...){
    va_list args;
    va_start(args, message);
    #ifdef LOGFILE
    FILE *ofp = fopen(LOGFILE,"a");
    if (ofp != NULL){
    		fprintf(ofp, "[%s]", __logging_timestr);
    		fprintf(ofp, "(%s:%d)", __logging_filename, __logging_linenumber);
    		fprintf(ofp," *** Fatal Error! ***: ");
    		vfprintf(ofp, message, args);
    		fprintf(ofp, "\n");
    		fclose(ofp);
    }else{
    	error(0, 0, "Failed to open %s for writing error message.\n", LOGFILE);
    }
    #endif
    //always print error messages to stdout
    printf("[%s]",__logging_timestr);
    printf("(%s:%d)", __logging_filename, __logging_linenumber);
    printf(" *** Fatal Error! ***: ");
    vprintf(message,args);
    printf("\n");
    va_end(args);
    exit(1);
}


#endif /*LOGGING_H_*/
