#include <sys/inotify.h>
#include <pthread.h>
#include <libconfig.h>

#include "logging.h"
#include "autofilename.h" //needed for the POSTSAVE_FILENAME macro
#include "led_control.h"
#include "connections_handler.h"

#define EVENT_SIZE (sizeof(struct inotify_event))
#define EVENT_BUF_LEN (2 *(EVENT_SIZE + 16))

//prototype static functions
static void getAFNFilename(char **filename);


/**********************************************************************************/
/**********************************************************************************/

int main(void)
{
#ifdef LOGFILE
	//give some message so that people know the program is working!
	printf("blur_serv: logging to file %s\n", LOGFILE);
#endif

	int length, i = 0;
	int fd;
	int wd;
	int j;
	char buffer[EVENT_BUF_LEN];
	char *filename = NULL;
	int n_events;
	char new_im_filename[100];
	FILE *ifp;
	
	connections_t con;
	led_ctrl_t led_ctrl;

	//initialise variables
	initialiseConnectionStruct(&con);
	initialiseLEDStruct(&led_ctrl);
		
	//start a new thread to handle client connections
	runConnectionsHandler(&con);
	
	//start a new thread to flash the LED
	runLEDIndicator(&led_ctrl);
	
	//read the filename we need to watch from the settings file
	getAFNFilename(&filename);
	DEBUG("AutoFileName post-save file = \'%s\'", filename);

	//check that this file actually exists - if it doesn't, then it is probably because
	//the main blur program hasn't started yet - sit and wait for it to appear
	while(access(filename, F_OK) != 0){
		DEBUG("Waiting for \'%s\' to be created.", filename);
	    sleep(1);
	}

	// create the INOTIFY instance
	fd = inotify_init();
	if ( fd < 0 ) {
		FATAL("Failed to initialise inotify");
	}

	//add the file to the watch list
	wd = inotify_add_watch( fd, filename, IN_CLOSE_WRITE );
	DEBUG("Watching \'%s\' for changes", filename);

	//enter an infinite loop of waiting for images to be written
	//signal(SIGINT, set_exit);
	while (1){
		//read to determine the event change happens on the file. Actually this read blocks until the change event occurs
		DEBUG("Reading inotify file");
		length = read( fd, buffer, EVENT_BUF_LEN );
		
		DEBUG("File write detected.");
		/*checking for error*/
		if (length < 0){
		FATAL("Read error on inotify instance.");
		}
		
		//read the list of change events
		n_events = 0;
		i = 0;
		while (i < length){
		  struct inotify_event *event = (struct inotify_event *) &buffer[i];
		  i += EVENT_SIZE + event->len;
		  n_events++;
		}
		
		//sanity check - we only expect to get one hit at a time
		if (n_events > 1){
		  WARN("Multiple events encountered - server is not keeping up!");
		}
		
		//flash the led to show that a new image has been captured
		signalNewImage(&led_ctrl);
		
		pthread_mutex_lock(&(con.mutex));
		if (con.count == 0){
		  //no clients connected - nothing to do
		  pthread_mutex_unlock(&(con.mutex));
		  continue;
		}
		pthread_mutex_unlock(&(con.mutex));
		
		//read the filename of the new image from the AFN post-save file
		ifp = fopen(filename, "r");
		if (ifp == NULL){
		  WARN("Failed to open %s for reading", filename);
		  continue;
		}
		
		if (fscanf(ifp,"%d\n", &i) != 1){ //skip the file number
		  DEBUG("Failed to read new image filename from file.");
		  fclose(ifp);
		  continue;
			
		}
		if (fscanf(ifp,"%100s", (char *)(&new_im_filename)) != 1){
		  DEBUG("Failed to read new image filename from file.");
		  fclose(ifp);
		  continue;
		}
		fclose(ifp);
		
		DEBUG("New image file \'%s\' detected", new_im_filename);
		
		//send the image data to all connected clients
		sendImageDataToClients(&con, new_im_filename);
	}
	/* Never get here */
	return 1;

}


/**********************************************************************************/
/**********************************************************************************/

//static function
void getAFNFilename(char **filename){
	//
	// Sets filename to the name of the AutoFileName post-save file (which is the one we need to
	// watch for new images)
	//
	config_t cfg;
	const char *base_dir_str;
	config_init(&cfg);
	if (config_read_file(&cfg, BLUR_CFG_FILE) != CONFIG_TRUE){
		FATAL("Error reading config file \'%s\' (line:%d):%s",BLUR_CFG_FILE, config_error_line(&cfg), config_error_text(&cfg));
	}

	//otherwise read the base_dir from the file
	if(config_lookup_string(&cfg, "capture.base_dir", &base_dir_str) != CONFIG_TRUE){
		FATAL("Error reading base_dir value from \'%s\':%s", BLUR_CFG_FILE, config_error_text(&cfg));
	}

	//make a copy of the filename string so that we can free the config memory
	int filename_len = strlen(base_dir_str) + strlen(POSTSAVE_FILENAME) + 2;

	*filename = malloc(sizeof(char) * filename_len);

	if(*filename != NULL){
		strcpy(*filename, base_dir_str);
		strcat(*filename, "/");
		strcat(*filename, POSTSAVE_FILENAME);
	}

	config_destroy(&cfg);

}

/**********************************************************************************/
/**********************************************************************************/
