// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.


#include"autofilename.h"

static int do_mkdir(const char *path, mode_t mode){
	// This function is stolen from J Leffler's mkpath
	// (C) JLSS 1990-91,1997-98,2001,2005,2008
    struct stat     st;
    int             status = 0;
    if (stat(path, &st) != 0)
    {
        /* Directory does not exist */
        if (mkdir(path, mode) != 0)
            status = -1;
    }
    else if (!S_ISDIR(st.st_mode))
    {
        errno = ENOTDIR;
        status = -1;
    }

    return(status);
}

int makeSubdirsForFile(const char *filename, mode_t mode){
	// Given the full path to a file, this function ensures that all
	// the directories in the path exist. If you just want to create a
	// directory structure, then make sure the path you pass in has a
	// terminating '/' e.g. makeSubdirsForFile("foo/bar/", 777), calling
	// makeSubdirsForFile("foo/bar", 777) will only create foo.
	//
	// Algorithm takes the pessimistic view and works top-down to ensure
	// each directory in path exists, rather than optimistically creating
	// the last element and working backwards.
	//
	// This function is a slightly modified version of J Leffler's mkpath
	// (C) JLSS 1990-91,1997-98,2001,2005,2008

    char *pp;
    char *sp;
    int status;
    char *copypath = strdup(filename);
    if (copypath == NULL){
    	WARN("Failed to create copy of filename string.");
    	return 1;
    }

    char *final_slash = strrchr(copypath,'/');
    if (final_slash == NULL){
    	WARN("filename does not appear to be a complete path name (contained no '/' characters). Giving up.");
    	free(copypath);
    	return 1;
    }

    *final_slash = '\0'; //ignore the name of the file
    status = 0;
    pp = copypath;
    while ((status == 0) && ((sp = strchr(pp, '/')) != 0))
    {
        if (sp != pp)
        {
            // Neither root nor double slash in path
            *sp = '\0';
            status = do_mkdir(copypath, mode);
            *sp = '/';
        }
        pp = sp + 1;
    }

    if (status == 0)
        status = do_mkdir(copypath, mode);

    free(copypath);
    return (status);
}


AutoFileName::AutoFileName(const char *base_dir, const int name_len, const int files_per_dir=500){
	// AutoFileName constructor:
	// base_dir - directory where you want the files to be saved
	// name_len - length of the filenames to be used, e.g. name_len=2 will produce files called 00, 01, 02 etc.
	// files_per_dir - maximum number of files you want to be put in any one subdirectory
	
	int base_dir_len = strlen(base_dir);
	current_filename = NULL;
	tmp_filename_buffer = NULL;
	this->files_per_dir = files_per_dir;

	num_saved = 0;

	//make copies of the output directory name
	this->base_dir = new char[base_dir_len + 1]; //+1 for \0
	strcpy(this->base_dir, base_dir);

	//ensure no trailing forward slash on base_dir name
	if (string_ends_with(this->base_dir, "/")){
		this->base_dir[base_dir_len - 1] = '\0';
	}

	//set the names of the persistance files
	pre_save_file = new char[base_dir_len + strlen(PRESAVE_FILENAME) + 2]; //+2 for '/' and '\0'
	sprintf(pre_save_file, "%s/%s", this->base_dir,PRESAVE_FILENAME);
	post_save_file = new char[base_dir_len + strlen(POSTSAVE_FILENAME) + 2]; //+2 for '/' and '\0'
	sprintf(post_save_file, "%s/%s", this->base_dir,POSTSAVE_FILENAME);

	initialise(name_len);
	
}

void AutoFileName::initialise(int name_len){
	// allocs the memory required for autofilenaming with the specified name length

	if (name_len < 1){
		WARN("Name length must be >=1. Changing it to 3.");
		name_len = 3;
	}

	if (name_len > 99){
		//in fact this test is a bit redundant given the ULONG overflow test
		//below - but you never know!
		FATAL("Insufficient string buffer size for name lengths > 99");
	}

	this->name_len = name_len;

	//free mem previously assigned for filename (it needs to be a different length now)
	delete[] current_filename;
	delete[] tmp_filename_buffer;
	current_filename = NULL;
	tmp_filename_buffer = NULL;

	//sanity check that name_len is not going to cause overflow
	if ((pow(10.0, (double)name_len)) > ULONG_MAX){
		FATAL("Number of possible files exceeds size of unsigned long - you can't have a name length this big!");
	}

	//calculate the max number of files that we can save with a filename of this length
	max_files = (unsigned long)(pow(10.0, (double)name_len));

	//how many folders are needed to meet to the files_per_dir requirement
	depth = (int)(log(max_files)/log(files_per_dir) + 0.5) - 1; //-1 because base_dir counts as one level of subdirs

	//alloc mem for complete path to file that we want to save
	current_filename = new char[strlen(base_dir) + ((depth + 1) * (name_len + 1)) + 11]; //+11 for filename extension of 10 chars and '\0'

	//alloc mem for temporary filename buffer
	tmp_filename_buffer = new char[name_len + 1]; //+1 for '\0'

	//create format string for subdir sprintf calls
	if (snprintf(subdir_fmt_str, SUBDIR_FMT_STR_LEN, "%%0%dlu", name_len) >= SUBDIR_FMT_STR_LEN){
		FATAL("subdir_fmt_str has been truncated. This should never happen! Has SUBDIR_FMT_STR_LEN been set to something weird in autofilename.h?");
	}

}

AutoFileName::~AutoFileName(void){
	delete[] base_dir;
	delete[] post_save_file;
	delete[] pre_save_file;
	delete[] current_filename;
	delete[] tmp_filename_buffer;
}

