// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.

/*
 * autofilename provides an AutoFileName class which allows saving of files with
 * automatically generated filenames. The idea is to ensure that old files are never
 * overwritten accidentally (for example if the computer clock gets messed up). It
 * is also designed to limit the number of files/folders created within a single folder.
 * Huge numbers of files in one folder often crash file browsers and are generally a pain.
 */

#ifndef AUTOFILENAME_H
#define AUTOFILENAME_H

#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <limits.h>

#include"logging.h"
#include"str_functions.h"

//define filenames where autofilename will save state. These will be created in the base directory.
#define PRESAVE_FILENAME "AFN_presave"
#define POSTSAVE_FILENAME "AFN_postsave"

//name_len should never be more than 2 digits long (therefore 7 is long
//enough for the format string), i.e. you will never want a filename 100 chars long.
#define SUBDIR_FMT_STR_LEN 7


#ifdef __cplusplus
//prototype function here since it is used in AutoFileName::save
int makeSubdirsForFile(const char *path, mode_t mode);


class AutoFileName
{
	public:
		AutoFileName(const char *base_dir, const int name_len, const int files_per_dir);
		~AutoFileName(void);
		template <typename T>
		int save(T *obj_to_save, const char *filename_extension);

	private:
		char * base_dir;
		char * pre_save_file;
		char * post_save_file;

		int name_len;
		int files_per_dir;
		unsigned long num_saved; //number of files saved to this base_dir
		int depth; //depth of dir structure needed to maintain files_per_dir
		unsigned long max_files; //max number of files that we can save with a filename of the given length

		char * current_filename;
		char * tmp_filename_buffer; //str manipulation memory space
		char subdir_fmt_str[SUBDIR_FMT_STR_LEN];

		void initialise(int name_len);

};

//need to define 'save' member function here since it is a template
template <typename T>
int AutoFileName::save(T *obj_to_save, const char * filename_extension){
	// for any object 'obj_to_save' saves it with an auto-generated filename and
	// the specified extension. obj_to_save needs to implement a member function
	// save(const char *filename).

	
	//check that extension is not longer than 10 chars (this is all we have allowed for in the memory allocation)
	if (strlen(filename_extension) > 10){
		FATAL("Filename extension must be 10 characters or less in length.");
	}

	unsigned long pre_save_num;
	unsigned long post_save_num;
	int status = 0;

	//read number of files saved from the persistance files.
	FILE *fp;
	if ((fp = fopen(pre_save_file,"r")) == NULL){
		DEBUG("No presave file found. Setting pre_save_num to 0.");
		pre_save_num = 0;
	} else {
		if (fscanf(fp,"%lu", &pre_save_num) != 1){
			WARN("Failed to read pre_save_num from %s", pre_save_file);
			pre_save_num = 0;
		}
		fclose(fp);
	}
	if ((fp = fopen(post_save_file,"r")) == NULL){
		DEBUG("No postsave file found. Setting post_save_num to 0.");
		post_save_num = 0;
	} else {
		if (fscanf(fp,"%lu", &post_save_num) != 1){
			WARN("Failed to read post_save_num from %s", post_save_file);
			post_save_num = 0;
		}
		fclose(fp);
	}

	//decide what our num_saved should be
	if (post_save_num > num_saved){
		//fresh instantiation of AFN - read num_saved from post save file
		num_saved = post_save_num;
	}

	if ((post_save_num == 0) && (pre_save_num > num_saved)){
		//couldn't read post save file, use presave file instead. Depending on circumstances
		//this could be greater than the actual number of files saved, but this isn't a major problem
		//this should only happen if the program fails to save a file, and then crashes. Hopefully this
		//won't happen too often so it should not be wildly out.
		num_saved = pre_save_num;
	}
	
	DEBUG("Current number saved = %d", num_saved);
	

	//build the full path for the file we are going to save
	unsigned long cur_num;
	do {
		//if the number saved is too big for the current name_length, then increase the name length
		while (num_saved >= max_files){
			WARN("Number of files saved has exceeded name length - increasing name length to %d", name_len + 1);
			initialise(name_len + 1);
		}

		strcpy(current_filename, base_dir);
		strcat(current_filename, "/");

		//create enough subdirs that files_per_dir can be met for full range of possible file names
		for(int i=0; i<depth; i++){
			cur_num = pow(files_per_dir, depth-i) * ((int)(num_saved / pow(files_per_dir, depth-i)));
			sprintf(tmp_filename_buffer, subdir_fmt_str, cur_num);
			strcat(current_filename, tmp_filename_buffer);
			strcat(current_filename,"/");
		}

		//now the filename (uses the same format string as the subdirs - zero padded to correct length)
		sprintf(tmp_filename_buffer, subdir_fmt_str, num_saved);
		strcat(current_filename, tmp_filename_buffer);
		strcat(current_filename, filename_extension);

		num_saved++;

	} while (access(current_filename, F_OK) == 0); //this is the final check to make sure the file doesn't already exist

	//save the state to the pre_save file
	if ((fp = fopen(pre_save_file, "w")) != NULL){
		fprintf(fp,"%lu", num_saved);
		fclose(fp);
	}
	
	DEBUG("Attempting to save to %s",current_filename);
	
	//build the required subdir structure
	if(makeSubdirsForFile(current_filename, 0777)){
		//we failed to build the subdir structure for some reason
		DEBUG("Failed to create directory structure");
		num_saved--; //we haven't saved a file with this number, so decrement it
		status = 1;
	} else {
		if (obj_to_save->save(current_filename)){
			num_saved--; //we haven't saved a file with this number, so decrement it
			status = 1;
		}
	}

	//save the state to the post_save file
	if ((fp = fopen(post_save_file, "w")) != NULL){
		fprintf(fp, "%lu\n%s", num_saved, current_filename);
		fclose(fp);
	}

	return status;

}

#endif //__cplusplus
#endif //AUTOFILENAME_H
