// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.

/*
 * camera.cpp
 *
 *  Created on: 30 Oct 2011
 *      Author: nialp
 */
#include"camera.h"
typedef struct {
	GMutex *mutex;

	guint16 packet_id;

	GSocket *socket;
	GSocketAddress	*interface_address;
	GSocketAddress	*device_address;

	GPollFD poll_in_event;

	void *buffer;

	unsigned int gvcp_n_retries;
	unsigned int gvcp_timeout_ms;

	gboolean is_controller;
} ArvGvDeviceIOData;
struct _ArvGvDevicePrivate {
	ArvGvDeviceIOData *io_data;

	void *heartbeat_thread;
	void *heartbeat_data;

	ArvGc *genicam;

	char *genicam_xml;
	size_t genicam_xml_size;
};
CameraBase::CameraBase(void){
	device = NULL;
	stream = NULL;
	genicam = NULL;
}

void CameraBase::onControlLost(void *x){
	WARN("Control Lost!");
}

int CameraBase::initCamera(void){

	DEBUG("Initialising Camera (CameraBase)");

	//char memory_buffer[100000];
	g_type_init();

	//wait for up to one minute to connect to the camera -
	// arv_open_device seems to take 1 second to complete, so no need to sleep
	int i=0;
	while (i<60){
		i++;
		device = arv_open_device("FLIR Systems-");
		if (device != NULL){
			break;
		}
	}
	if (device == NULL){
		WARN("Failed to open the device.");
		return 1;
	}

	g_signal_connect(device, "control-lost",GCallback(CameraBase::onControlLost),NULL);

	if ((genicam = arv_device_get_genicam(device)) == NULL){
		WARN("Failed to get the GeniCam interface.");
		return 1;
	}

	if ((stream = arv_device_create_stream(device, NULL, NULL))==NULL){
		WARN("Failed to create the stream.");
		return 1;
	}

	return 0;

}



CameraBase::~CameraBase(void){

	DEBUG("Calling CameraBase destructor\n");

	//free memory - note that the buffer(s) are released by the stream
	if(stream != NULL){
		g_object_unref(stream);
	}

	if(device != NULL){
		g_object_unref(device);
	}

	arv_shutdown();
}


int CameraBase::setConfig(const char* name, bool val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (!ARV_IS_GC_BOOLEAN(node)){
		WARN("Cannot set config \'%s\' to %d. Config is not an integer.", name, val);
		return 1;
	}

	arv_gc_boolean_set_value(ARV_GC_BOOLEAN(node), val, &err);

	if (err != NULL){
		WARN("Failed to set config \'%s\' to %d. %s", name, val, err->message);
		g_error_free(err);
		return 1;
	}
	return 0;
}


int CameraBase::setConfig(const char* name, long val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//integer nodes can also be enumerations, which have to be dealt with differently
	if (ARV_IS_GC_INTEGER(node)){
		arv_gc_integer_set_value(ARV_GC_INTEGER(node), (gint64)val, &err);
		if (err != NULL){
			goto fail;
		}

	} else if (ARV_IS_GC_ENUMERATION(node)){
		//if it is an enumeration, then we should check that the value we are trying to set
		//it to is a valid entry
		guint n;
		gint64 *valid_values = arv_gc_enumeration_get_available_int_values(ARV_GC_ENUMERATION(node), &n, &err);
		if (err != NULL){
			goto fail;
		}
		for(unsigned int i=0; i<n; i++){
			if(valid_values[i] == val){
				arv_gc_enumeration_set_int_value(ARV_GC_ENUMERATION(node), (gint64)val, &err);
				if (err != NULL){
					goto fail;
				}
				return 0;
			}
		}
		//make a list of possible values that it could have had
		std::stringstream ss;
		ss << "[";
		if(n > 0){
			for(unsigned int i=0; i<n-1; i++){
				ss << valid_values[i] << ", ";
			}
			ss << valid_values[n-1] << "]";
		}

		WARN("Cannot set config \'%s\' to %d. Valid values for \'%s\' are: %s", name, val, name, ss.str().c_str());
		return 1;
	} else {
		WARN("Cannot set config \'%s\' to %d. Config does not have an integer type.", name, val);
		return 1;
	}

	return 0;

	fail:
		if(err != NULL){
			WARN("Cannot set config \'%s\' to %d. %s", name, val, err->message);
			g_error_free(err);
		}
		return 1;
}



int CameraBase::setConfig(const char* name, double val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (!ARV_IS_GC_FLOAT(node)){
		WARN("Cannot set config \'%s\' to %lf. Config is not a float.", name, val);
		return 1;
	}

	arv_gc_float_set_value(ARV_GC_FLOAT(node), val, &err);

	if(err != NULL){
		WARN("Cannot set config \'%s\' to %lf. %s", name, val, err->message);
		g_error_free(err);
		return 1;
	}

	return 0;
}



int CameraBase::setConfig(const char* name, const char* val){

	ArvGcNode *node = arv_gc_get_node(genicam, name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//string nodes can also be enumerations, which have to be dealt with differently
	if (ARV_IS_GC_STRING(node)){
		arv_gc_string_set_value(ARV_GC_STRING(node), val, &err);
		if (err != NULL){
			goto fail;
		}
	} else if(ARV_IS_GC_ENUMERATION(node)){
		//if it is an enumeration, then we should check that the value we are trying to set
		//it to is a valid entry
		const GSList *iter;

		for (iter=ARV_GC_ENUMERATION(node)->entries; iter != NULL; iter = iter->next){
			if (g_strcmp0(arv_gc_feature_node_get_name((ArvGcFeatureNode*)(iter->data)), val) == 0) {
				arv_gc_enumeration_set_string_value(ARV_GC_ENUMERATION(node), val, &err);
				if (err != NULL){
					goto fail;
				}
				return 0 ;
			}
		}

		//make a list of possible values that it could have had
		std::stringstream ss;
		ss << "[";
		for (iter = ARV_GC_ENUMERATION(node)->entries; iter != NULL; iter = iter->next){
				ss << arv_gc_feature_node_get_name((ArvGcFeatureNode*)(iter->data)) << ", ";
			}
		ss << "\b\b]";

		WARN("Cannot set config \'%s\' to %s. Valid values for \'%s\' are: %s", name, val, name, ss.str().c_str());

		return 1;
	} else{
		WARN("Cannot set config \'%s\' to %s. Config is not a string.", name, val);
		return 1;
	}

	return 0;

	fail:
		if(err != NULL){
			WARN("Cannot set config \'%s\' to %s. %s", name, val, err->message);
			g_error_free(err);
		}
		return 1;

}


int CameraBase::getConfig(const char* name, bool &val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);

	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (!ARV_IS_GC_BOOLEAN(node)){
		WARN("Node \'%s\' is not a boolean!", name);
		return 1;
	}

	val = (bool)arv_gc_boolean_get_value(ARV_GC_BOOLEAN(node), &err);

	if(err != NULL){
		WARN("Cannot get value of config \'%s\'. %s", name, err->message);
		g_error_free(err);
		return 1;
	}
	return 0;
}



int CameraBase::getConfig(const char* name, long &val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (ARV_IS_GC_INTEGER(node)){
		val = (long)arv_gc_integer_get_value(ARV_GC_INTEGER(node), &err);
	} else if (ARV_IS_GC_ENUMERATION(node)){
		val = (long)arv_gc_enumeration_get_int_value(ARV_GC_ENUMERATION(node), &err);
	} else {
		WARN("Node \'%s\' is not an integer!", name);
		return 1;
	}

	if(err != NULL){
		WARN("Cannot get value of config \'%s\'. %s", name, err->message);
		g_error_free(err);
		return 1;
	}

	return 0;
}



int CameraBase::getConfig(const char* name, float &val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (!ARV_IS_GC_FLOAT(node)){
		WARN("Node \'%s\' is not a float!", name);
		return 1;
	}

	val = (float)arv_gc_float_get_value(ARV_GC_FLOAT(node), &err);

	if(err != NULL){
		WARN("Cannot get value of config \'%s\'. %s", name, err->message);
		g_error_free(err);
		return 1;
	}

	return 0;
}



int CameraBase::getConfig(const char* name, char** val){

	ArvGcNode *node = arv_gc_get_node(genicam,name);
	GError *err = NULL;

	if (node == NULL){
		WARN("Failed to get node \"%s\"",name);
		return 1;
	}

	//check that the node is of the correct type
	if (ARV_IS_GC_STRING(node)){
		const char* tmp = arv_gc_string_get_value(ARV_GC_STRING(node), &err);
		if ((tmp != NULL) && (err == NULL)){
			*val = strdup(tmp);
			if (*val == NULL){
				WARN("Failed to make a copy of the string for config \'%s\'.",name);
				return 1;
			}
			return 0;
		}
	} else if (ARV_IS_GC_ENUMERATION(node)){
		const char* tmp = arv_gc_enumeration_get_string_value(ARV_GC_ENUMERATION(node), &err);
		if ((tmp != NULL) && (err == NULL)){
			*val = strdup(tmp);
			if (*val == NULL){
				WARN("Failed to make a copy of the string for config \'%s\'.",name);
				return 1;
			}

			return 0;
		}
	} else {
		WARN("Node \'%s\' is not a string!", name);
		return 1;
	}

	if(err != NULL){
		WARN("Cannot get value of config \'%s\'. %s", name, err->message);
		g_error_free(err);
	} else {
		WARN("aravis get string value returned NULL without setting an error!");
	}

	return 1;
}


SimpleCamera::~SimpleCamera(void){
	delete image;
}


//static method
SimpleCamera * SimpleCamera::createCamera(bool enable_fast_compress){

	SimpleCamera *c = new SimpleCamera;
	ImageBase *im = new ImageBase(enable_fast_compress);
	if(c->initCamera(im)){
		//failed to initialise the camera
		delete c;
		return NULL;
	}
	return c;
}


int SimpleCamera::initCamera(ImageBase *im){

	//do this first so that if anything else fails, then the image still gets deleted in the destructor
	image = im;

	//call base class initialisation function
	if( this->CameraBase::initCamera()){
		return 1;
	}

	DEBUG("Initialising Camera (SimpleCamera)");

	ArvBuffer *buffer;
	capture_node = arv_gc_get_node (genicam, "AcquisitionStart");


	//find the size required for the image buffer
	long buffer_size;
	if (getConfig("PayloadSize", buffer_size)){
		return 1;
	}

	//create a buffer of the desired size
	if((buffer = arv_buffer_new(buffer_size, NULL)) == NULL){
		WARN("Failed to create buffer");
		return 1;
	}

	DEBUG("Created single buffer of size %d",buffer_size);

	//put the buffer into the stream and set the flag to show that there is a
	//buffer in the stream
	arv_stream_push_buffer(stream, buffer);
	buffer_flag = true;

	//set the camera up for single frame capture
	if(setConfig("AcquisitionMode", "SingleFrame")){
		WARN("Failed to set the AcquisitionMode.");
		return 1;
	}

	return 0;

}



void SimpleCamera::recycleImage(ImageBase *im){

	arv_stream_push_buffer(stream, im->popBuffer());
	buffer_flag = true;
}

int SimpleCamera::reInitialise(void){
	//this is a quick hack to try and fix the loss of control of the camera (which
	//seems to happen every 10 mins or so) here we just try to reopen the device
	//free memory - note that the buffer(s) are released by the stream
	if(stream != NULL){
		g_object_unref(stream);
	}

	if(device != NULL){
		g_object_unref(device);
	}

	arv_shutdown();

	//now reopen the device
	return initCamera((ImageBase*)image);


}

ImageBase * SimpleCamera::captureImage(void){

	if (!buffer_flag){
		WARN("Attempt to capture image before the buffer has been reset");
		return NULL;
	}

	ArvBuffer *tmp_buffer;
	GError *err = NULL, *err2 = NULL, *err3 = NULL;

	//take the image
	arv_gc_command_execute(ARV_GC_COMMAND(capture_node), &err);
	if (err != NULL){
		WARN("Failed to capture image. %s",err->message);
		g_error_free(err);
		return NULL;
	}/*
		//check and see if we have lost control of the camera
		guint32 value;
		arv_device_read_register (device, ARV_GVBS_CONTROL_CHANNEL_PRIVILEGE_OFFSET, &value, &err2);
		if(err2 != NULL){
			WARN("Failed to read control register (assuming loss of control). %s",err2->message);
			g_error_free(err2);
			value=0;
		}
		if ((value & (ARV_GVBS_CONTROL_CHANNEL_PRIVILEGE_CONTROL | ARV_GVBS_CONTROL_CHANNEL_PRIVILEGE_EXCLUSIVE)) == 0){
			//we have lost control - attempt to reinitialise
			if (!reInitialise()){
				WARN("Failed to reinitialise camera.");
				return NULL;
			}else{
				arv_gc_command_execute(ARV_GC_COMMAND(capture_node), &err3);
				if (err3 != NULL){
					WARN("Failed to capture image with reinitialised camera. %s",err3->message);
					g_error_free(err3);
					return NULL;
				}
			}

		} else {
			//we still have control, something else has gone wrong
			return NULL;
		}
	}*/

	//wait for up to 2 seconds for the buffer to appear in the output queue.
	//note that if the timeout is reached then pop_buffer doesn't get called again -
	//so if wait_loops !<2000 then the buffer will ALWAYS be left in the stream queue.
	int wait_loops = 0;
	while ((wait_loops < 2000) && ((tmp_buffer = arv_stream_pop_buffer(stream)) == NULL)){
		wait_loops ++;
		g_usleep (1000); //sleep for 1 millisecond
	}
	buffer_flag = false;
	DEBUG("Waited %d milliseconds seconds for image", wait_loops);

	//record the capture time
	image->setCaptureTimeAsNow();

	//if the timeout was reached, then we have a problem. Is the buffer in the output
	//queue, or did the camera totally fail to take an image (in which case the buffer
	//is still sat in the input queue)?
	if (! (wait_loops < 2000)){
		WARN("Image capture timeout exceeded.");
		if(tmp_buffer != NULL){
			WARN("Capture timeout exceeded, but buffer is not NULL - something has gone wrong!");
			arv_stream_push_buffer(stream, tmp_buffer);
			buffer_flag = true;
			return NULL;
		}

		//try to pop the buffer from both the input and output queue
		if ((tmp_buffer = arv_stream_pop_buffer(stream))!=NULL){
			arv_stream_push_buffer(stream, tmp_buffer);
			buffer_flag = true;
		}else if ((tmp_buffer = arv_stream_pop_input_buffer(stream))!=NULL){
			arv_stream_push_buffer(stream, tmp_buffer);
			buffer_flag = true;
		}else{
			WARN("Buffer is neither in input queue nor output queue - where has it gone?!");
			//note that buffer does not get reset, so subsequent image captures will fail - this
			//is fine because eventually the main blur program will re-instantiate the camera. This
			//probably leaks memory, but if we can't find the buffer to free it, then there is no
			//choice.
		}
		return NULL;
	}

	if (tmp_buffer->status == ARV_BUFFER_STATUS_SUCCESS){
		DEBUG("Buffer popped from queue contains valid image of size %dx%d.", tmp_buffer->width, tmp_buffer->height);
		//hooray! everything worked.
		//set the image buffer to that just captured
		image->setBuffer(tmp_buffer);
		return image;
	}

	//if the image capture was not successful (buffer does not contain a valid image
	//for whatever reason), then reset the buffer and return NULL
	WARN("Buffer did not contain a valid image.");
	arv_stream_push_buffer(stream, tmp_buffer);
	buffer_flag = true;

	return NULL;
}


