// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.

/*
 * image.cpp
 *
 *  Created on: 3 Nov 2011
 *      Author: nialp
 */
#include"image.h"

template<> void ImageBase::setInfoField<double>(const char *name, double value){
	//specialization for double values, the standard template records them
	//in the image header in the form 3.14e9, which is no good since you lose
	//half of the information - so this function forces fixed width on the stringstream
	std::stringstream ss;

	ss << std::fixed << value;

	//save in map
	info_fields[name] = ss.str();
}


void ImageBase::setCaptureTimeAsNow(void){

	struct timeval tv;
	gettimeofday(&tv,NULL);
	//convert the time to a float and store it in the ImageBase object
	capture_time = (double)(tv.tv_sec) + ((double)(tv.tv_usec))/1000000.0;

	//set this as a text field to be stored in the file
	setInfoField("capture_time", capture_time);
}



int ImageBase::save(const char *filename){

	if (buffer == NULL){
		WARN("Cannot save image with NULL buffer.");
		return 1;
	}

	if ((buffer->width < 5) || (buffer->height < 5)){
		WARN("Image dimension is <5 pixels. Abandoning save.");
		return 1;
	}

	if (preProcess()){
		WARN("Pre-processing failed when saving image %s", filename);
		return 1;
	}

	if (string_ends_with(filename,".png")){
		return saveAsPNG(filename);
	} else {
		WARN("Unknown file extension for image file %s. Giving up.", filename);
		return 1;
	}
}


int ImageBase::saveAsPNG(const char* filename){

	FILE *ofp = NULL;
	char filename_and_ext[200];

	//check that the filename has the correct extension
	if(strlen(filename)>196){
		FATAL("Filename is too long! Must be less than 196 characters.");
	}

	if(string_ends_with(filename,".png")){
		sprintf(filename_and_ext,"%s",filename);
	}else{
		sprintf(filename_and_ext,"%s.png",filename);
	}

	//try to open the file for writing
	if ((ofp = fopen(filename_and_ext,"wb")) == NULL){
		WARN("Saving image failed. Cannot open file for writing.");
		return 1;
	}

	// Create the png structure
	png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (!png_ptr){
		WARN("Saving image failed. Cannot create png_struct.");
		fclose(ofp);
		return 1;
	}

	// Create the png info structure
	png_infop info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr){
		WARN("Saving image failed. Cannot create png_info structure.");
		png_destroy_write_struct(&png_ptr,(png_infopp) NULL);
		fclose(ofp);
		return 1;
	}

	// setup error handling for libpng
	if (setjmp(png_jmpbuf(png_ptr))){
		WARN("Saving image failed. libpng encountered a problem");
		png_destroy_write_struct(&png_ptr, &info_ptr);
		fclose(ofp);
		return 1;
	}

	// initialise the IO
	png_init_io(png_ptr, ofp);

	int bit_depth;

	if(buffer->pixel_format == ARV_PIXEL_FORMAT_MONO_16){
		bit_depth = 16;
	}else if (buffer->pixel_format == ARV_PIXEL_FORMAT_MONO_8){
		bit_depth = 8;
	}else{
		FATAL("Can't process this pixel format - %x",buffer->pixel_format);
	}

	// fill in the png info struct
	png_set_IHDR(png_ptr, info_ptr, getWidth(), getHeight(), bit_depth, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_color_8 c;
	c.gray = bit_depth;
	png_set_sBIT(png_ptr, info_ptr, &c);
	DEBUG("Image sBit = %d",bit_depth);

	if (fast_compress){
		DEBUG("Fast compress is set - using best speed.");
		png_set_compression_level(png_ptr, Z_BEST_SPEED);
	} else {
		DEBUG("Fast compress is not set - using best compression.");
		png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	}

	int k;
	png_bytep *row_pointers = new png_bytep[getHeight()];

	for (k = 0; k < getHeight(); k++){
	 row_pointers[k] = ((png_byte*)(buffer->data)) + k*getWidth()*(bit_depth / 8);
	}

	//fill in all the text fields containing the image metadata
	fillInTextFields(png_ptr,info_ptr);

	png_set_rows(png_ptr, info_ptr, row_pointers);

	//actually write the image to file - aravis seems to do an endianess conversion
	//so we need to convert back to big-endian before writing to png, we also rescale
	//all the colour values to fill the pixel depth
	png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_SWAP_ENDIAN, NULL);

	// Clean up after the write, and free any memory allocated
	delete[] row_pointers;
	row_pointers = NULL;

	png_destroy_write_struct(&png_ptr, &info_ptr);

	// Close the file
	fclose(ofp);

	DEBUG("Image saved as %s",filename);

	return 0;
}


void ImageBase::fillInTextFields(png_structp png_ptr, png_infop info_ptr){

	int num_of_entries = (int)(info_fields.size());

	png_text *text_ptr = new png_text[num_of_entries];

	std::map<const char*, std::string>::iterator it;

	it = info_fields.begin();
	for(int i=0; i<num_of_entries; i++){
		//note that libpng deep copies the the png_text struct, so it really is read-only and therefore it is safe to cast away the const'ness of these pointers
		text_ptr[i].key = const_cast<char*>((*it).first);
		text_ptr[i].text = const_cast<char*>((*it).second.c_str());
		text_ptr[i].compression = PNG_TEXT_COMPRESSION_NONE;
		text_ptr[i].text_length = strlen(text_ptr[i].text);
		it++;
	}

	png_set_text(png_ptr, info_ptr, text_ptr, num_of_entries);

	delete[] text_ptr;
}

