// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.


/*
 * image.hpp
 *
 *  Created on: 31 Oct 2011
 *      Author: nialp
 */

#ifndef IMAGE_HPP_
#define IMAGE_HPP_

#include <stdlib.h>
#include <png.h>
#include <time.h>
#include <sys/time.h>
#include <limits.h>
#include <map>
#include <string>
#include <sstream>
#include <zlib.h>

extern "C" {
#include <arv.h>
}

#include "str_functions.h"
#include "logging.h"

//double check that sizeof unsigned short is 2 (16 bits) - we are relying on this when
//we save the image.
#if USHRT_MAX != 65535
#error "Require a 2byte unsigned short implementation for direct saving of image buffers to PNG"
#endif


class ImageBase{

	public:
		virtual ~ImageBase(void){};
		ImageBase(bool enable_fast_compress){buffer = NULL; fast_compress = enable_fast_compress;};
		virtual int save(const char *filename);
		void setCaptureTimeAsNow(void);
		template<typename T> void setInfoField(const char *name, T value);
		inline double getCaptureTime(void){return capture_time;};
		inline int getWidth(void){if(buffer != NULL){return buffer->width;} else {return 0;}};
		inline int getHeight(void){if(buffer != NULL){return buffer->height;} else {return 0;}};
		inline ArvBuffer * popBuffer(void){ArvBuffer *ptr = buffer; buffer=NULL; return ptr;};
		inline void setBuffer(ArvBuffer *buffer){if (this->buffer == NULL){this->buffer = buffer;}else{FATAL("Attempt to overwrite non-NULL image buffer - this will cause memory leaks.");}};
		inline int getPixelFormat(void){if(buffer != NULL){return buffer->pixel_format;} else {return 0;}};

	protected:
		virtual int preProcess(void){return 0;};
		int saveAsPNG(const char* filename);
		ArvBuffer *buffer;
	private:

		std::map<const char *, std::string> info_fields; //map containing name,value pairs of metadata to be written to the image file when it is saved
		double capture_time; //seconds since the epoch, accurate to the nearest millisecond
		bool fast_compress;

		void fillInTextFields(png_structp png_ptr, png_infop info_ptr);
};


template<typename T>
void ImageBase::setInfoField(const char *name, T value){

	//convert value to str
	std::stringstream ss;

	ss << value;

	//save in map
	info_fields[name] = ss.str();

}


#endif /* IMAGE_HPP_ */
