// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.


#include <libconfig.h++>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <unistd.h>

#include "sc645.h"
#include "autofilename.h"
#include "logging.h"
#include "image.h"


static bool stay_alive = true; //global variable to control program exit
static int failed_capture_loops = 0; //global variable to count the number of

static void set_exit(int signal){
	//signal handler function for keyboard interrupt

	LOGMESSAGE("blur received exit signal (keyboard interrupt). Exiting.");

	//setting stay_alive = false will break the main capture loop (in main())
	// when it has finished taking/saving the current image.
	stay_alive = false;
}


static int runCapture(Setting &capture_cfg, Setting &camera_cfg){

	int failed_captures = 0;
	int refocus_counter = 0;
	int get_time_status = 0;
	struct timeval t1, t2;
	double time_taken;
	ImageBase *im;

	//make local copies of the configs we are going to use - this way if there is
	//a problem with the config file (or with the type of one of the configs) it is
	//found out immediately rather than when the first image is captured.
	try
	{
		const float capture_interval = capture_cfg["capture_interval"];
		const char *base_dir = capture_cfg["base_dir"];
		const int name_len = capture_cfg["filename_length"];
		const int files_per_dir = capture_cfg["files_per_dir"];
		const int focus_position = capture_cfg["focus_position"];
		const char *filename_extension = capture_cfg["filename_extension"];
		const int capture_fail_threshold = capture_cfg["capture_fail_threshold"];
		const int refocus_interval = capture_cfg["refocus_interval"];
		const bool fast_compress = capture_cfg["fast_compress"];
		const int low_T_threshold = capture_cfg["low_T_threshold"];

		AutoFileName afn = AutoFileName(base_dir, name_len, files_per_dir);
		SC645Camera *camera = SC645Camera::createCamera(camera_cfg, capture_cfg);

		if(camera == NULL){
			return 1;
		}

		//set the focus position of the camera - note that this is not done using dynamic
		//camera settings because the camera has to be refocused periodically (for some reason
		//the focus position seems to drift slightly).
		camera->setFocusPos(focus_position);

		//now that it is focused, redo the NUC correction otherwise we have to wait for the
		//autoNUC to happen which might take a while and the images will be messed up in the meantime
		camera->performNUC();
		
		//if the refocus interval has been set, then disable the autoNUC correction. This is desireable
		//so that images are not being captured when the NUC happens. However, no NUC corrections ever 
		//might damage the camera, so if the interval has been set to -1, then we enable autoNUC
		if (refocus_interval > 0){
			camera->setConfig("NUCMode", "Off");
		} else{
			camera->setConfig("NUCMode", "Automatic");
		}

		//enter main capture loop - runs until interrupted
		signal(SIGINT, set_exit);//set the exit signal handler
		while (stay_alive){
			get_time_status = gettimeofday(&t1, NULL);

			im = (ImageBase*)(camera->captureImage());

			if (im == NULL){
				//capture failed for some reason
				WARN("Failed to capture image.");
				failed_captures++;
			} else if (afn.save(im, filename_extension)){
				//afn.save calls the preprocessing function too.
				WARN("Failed to save image.");
				failed_captures++;
			} else {
				//only interested in recurring failures - so since we managed to save
				//an image, reset the counters
				failed_captures = 0;
				failed_capture_loops = 0;
			}

			//recycle the image buffer for the next capture attempt
			if (im != NULL){
				camera->recycleImage((IRImage*)im);
			}

			if (failed_captures >= capture_fail_threshold){
				//start again
				WARN("Capture fail threshold exceeded. Restarting capture loop.");
				delete camera;
				signal(SIGINT, SIG_DFL); //unset the signal handler
				return 1;
			}
			if (refocus_interval > 0){
				refocus_counter++;
				if (refocus_counter > refocus_interval){
					camera->refocus();
					camera->performNUC();
					refocus_counter = 1;
				}
			}

			get_time_status += gettimeofday(&t2, NULL);

			if (get_time_status == 0){
				time_taken = (t2.tv_sec - t1.tv_sec) + ((t2.tv_usec - t1.tv_usec)/1000000.0);
				DEBUG("Took %0.3lf seconds to capture and store image.", time_taken);
			} else {
				time_taken = 0;
				WARN("gettimeofday() failed. Assuming a capture/save time of 0 seconds.");
			}

			if(time_taken < capture_interval){
				usleep((useconds_t)(((capture_interval - time_taken)*1000000) + 0.5));
			}
		}

		delete camera;
	}
	catch (SettingTypeException &ex)
	{
		FATAL("Error getting config \'%s\' from %s. Incorrect data type.", ex.getPath(), BLUR_CFG_FILE);
	}
	catch (SettingNotFoundException &ex)
	{
		FATAL("Error getting config \'%s\' from %s. Config does not exist.", ex.getPath(), BLUR_CFG_FILE);
	}

	signal(SIGINT, SIG_DFL); //unset the signal handler
	return 0;
}


int main(void){
	Config cfg;
#ifdef LOGFILE
	//give some message so that people know the program is working!
	printf("blur: logging to file %s\n", LOGFILE);
#endif
	LOGMESSAGE("----- PROGRAM RESTART -----");

	// see if the config file exists and if not, does the template file exist.
	if (access(BLUR_CFG_FILE, F_OK) != 0){
		//see if the template file exists
		if (access(BLUR_CFG_FILE ".template", F_OK) != 0){
			FATAL("Config file %s does not exist and neither does %s. Did blur install correctly?", BLUR_CFG_FILE, BLUR_CFG_FILE".template");
		} else {
			FATAL("Please edit the template config file %s and resave it as %s.",BLUR_CFG_FILE".template",BLUR_CFG_FILE);
		}
	}


	//Read the config file. If there is an error, report it and exit.
	try
	{
		cfg.readFile(BLUR_CFG_FILE);
	}
	catch(const FileIOException &fioex)
	{
		FATAL("I/O error trying to read config file \'%s\'.", BLUR_CFG_FILE);
	}
	catch(ParseException &pex)
	{
		FATAL("Error on line %d of the config file: %s", pex.getLine(), pex.getError());
	}

	//if these fail then the program will exit - this is fine as there isn't anything we can do about it.
	try
	{
		Setting &capture_cfg = cfg.lookup("capture");
		Setting &camera_cfg = cfg.lookup("camera");

		//enter infinite loop of capturing and saving images
		while(runCapture(capture_cfg, camera_cfg)){
			if (failed_capture_loops > 3){
				//after three totally failed capture loops (i.e no images successfully saved)
				//we begin to suspect that something is badly wrong....restart the whole system
				//and hope for the best (note that we rely on the blur_init script to reboot the
				//system)

				WARN("Capture loop failure threshold exceeded. Exiting.");
				return 2;
			}
			failed_capture_loops++;
			LOGMESSAGE("runCapture() function encountered an error - retrying in 2 seconds");
			sleep(2);
		}
	}
	catch (SettingNotFoundException &ex)
	{
		FATAL("Failed to read %s configs from %s.", ex.getPath(), BLUR_CFG_FILE);
	}

	return 0;
}
