// Copyright (C) Nial Peters 2011
//
// This file is part of blur.
// blur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// blur is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with blur.  If not, see <http://www.gnu.org/licenses/>.

#include "sc645.h"

void IRImage::getMinMax(uint16_t &min_pix, uint16_t &max_pix){
	//gets the minimum and maximum pixel values from the image, ignoring any
	//bad pixels.
	//In the interests of efficiency, this function does not look at the edge pixels
	//of the image. That way each pixel analysed always has 8 neighbours and processing
	//is simpler.
	double sum = 0;
	double sum_sq = 0;
	const long num_of_pix = buffer->width * buffer->height;
	int j;
	int bad_relations;

	//bad pixel threshold is 3 times the standard deviation of the whole image - calculate it!
	for (long i=0; i<num_of_pix; i++){
		sum += ((uint16_t*)(buffer->data))[i];
		sum_sq += ((uint16_t*)(buffer->data))[i] * ((uint16_t*)(buffer->data))[i];
	}

	double threshold = 3.0 * sqrt((sum_sq/num_of_pix) - ((sum/num_of_pix)*(sum/num_of_pix)));

	uint16_t *ptr = (uint16_t*)(buffer->data) + buffer->width + 1; //skip first row and first column
	uint16_t * neighbours[] = {ptr - buffer->width - 1,\
							   ptr - buffer->width,\
							   ptr - buffer->width + 1,\
							   ptr - 1,\
							   ptr + 1,\
							   ptr + buffer->width - 1,\
							   ptr + buffer->width,\
							   ptr + buffer->width + 1}; //the eight nearest neighbours of ptr


	//initialise the min and max values to the value of the first pixel
	max_pix = *ptr;
	min_pix = *ptr;

    //std::stringstream bad_pix_str;
	//bad_pix_str << "[";

	for (unsigned int y=1; y < buffer->height - 1; y++){
		for (unsigned int x=0; x <buffer->width - 2; x++){
			bad_relations = 0;

			for (j=0; j<8; j++){
				if (fabs(*(neighbours[j]) - *ptr) > threshold){
					bad_relations++;
				}

				//update the neighbour ready for the next iteration
				neighbours[j] += 1;
			}

			//if the threshold is reached for the majority of neighbours - then claim it is a bad pixel
			if ((bad_relations / 8.0) < 0.5){
				//the pixel is good and should be included in the min/max calculation
				if (*ptr > max_pix){
					max_pix = *ptr;
				}else if (*ptr < min_pix){
					min_pix = *ptr;
				}

			} else {
				//pixel is potentially bad - ignore it.
				//bad_pix_str << "("<< x << "," << y << ")";
				DEBUG("Found bad pixel at %d,%d.", x,y);
			}

			//move to the next pixel in the row
			ptr += 1;
		}

		//move to the second column of the next row
		ptr += 2; //ptr has already been incremented by one during the inner loop

		j = 0;
		while (j<8){
			neighbours[j] += 2; //neighbours have also already been incremented in inner loop
			j++;
		}
	}
	//bad_pix_str << "]";
	//setInfoField("bad_pixels", bad_pix_str.str());
}

void IRImage::convertTo8bit(int min_pix, int max_pix){
	const long num_of_pix = buffer->width * buffer->height;
	const float scaling_factor = 255.0/(max_pix - min_pix);

	if (max_pix > min_pix){
		//convert the data to 8 bit - scaling it and overwriting the old 16bit data as we go.
		//this is done to conserve memory (rather than copying data into a new buffer)
		for (long i=0; i<num_of_pix; i++){
			if(((uint16_t *)buffer->data)[i] < min_pix){
				//if min_pix has been set to a low_T threshold, then some of the pixels might have lower values
				//set these to 0
				((char *)buffer->data)[i] = 0;
			} else if (((uint16_t *)buffer->data)[i] > max_pix){
				//bad pixels that have been excluded from the min/max calculation may have a value
				//greater than max_pix. Set these to 255
				((char *)buffer->data)[i] = 255;
			} else {
				((char *)buffer->data)[i] = (char)((((((uint16_t *)buffer->data)[i] - min_pix) * scaling_factor) + 0.5));
			}
		}
	}else{
		//set the whole image to black
		memset((char *)buffer->data, 0, num_of_pix);
	}

	//record that we have changed the bit depth of the buffer
	buffer->pixel_format = ARV_PIXEL_FORMAT_MONO_8;
	setInfoField("Converted_to_8bit","True");

}

int IRImage::preProcess(void){

	// double check that the image buffer has the expected bit depth
	if(buffer->pixel_format != ARV_PIXEL_FORMAT_MONO_16){
		// this should never happen, since it is checked for by SC645::captureImage,
		// but just in case...
		WARN("Pixels in buffer did not have expected bit depth (uint16).");
		return 1;
	}

	// double check that the sizes all look good (2 is the size of uint16)
	if (2 * buffer->width * buffer->height != buffer->size){
		WARN("Buffer size and/or data-type information appears to be corrupted.");
		return 1;
	}

	//get the min and max pixel values
	uint16_t max_pix, min_pix;
	getMinMax(min_pix, max_pix);
	
	//if the minimum pixel value is less than the low Temperature threshold, then update the value
	if(min_pix < low_T_threshold){
		min_pix = low_T_threshold;
	}
	
	//Note that if the max pixel value is less than the min pixel value, the whole image will
	//be set to black by the 8bit conversion.
	
	setInfoField("max_pixel_value", max_pix);
	setInfoField("min_pixel_value", min_pix);
	setInfoField("low_T_threshold", low_T_threshold);
	
	convertTo8bit(min_pix, max_pix);

	return 0;
}


//static memeber function
SC645Camera * SC645Camera::createCamera(Setting &camera_cfg, Setting &capture_cfg){

	SC645Camera *c = new SC645Camera;
	IRImage *im = new IRImage(capture_cfg["fast_compress"], capture_cfg["low_T_threshold"]); //this will be deleted by ~SimpleCamera
	if(c->initCamera(im, camera_cfg)){
		//failed to initialise the camera
		delete c;
		return NULL;
	}
	return c;
}


int SC645Camera::initCamera(IRImage *im, Setting &camera_cfg){

	image = im; //note that im gets destructed by ~SimpleCamera, because that also stores a pointer to it

	//call base class initialisation function
	if(this->SimpleCamera::initCamera((ImageBase *)im)){
		return 1;
	}

	//set all the camera configs from the values that were stored in the config file,
	//and copy these into the image metadata.

	//start with the static configs - these just need to be put into the image header data
	Setting &static_configs = camera_cfg["static"];
	for (int i=0; i< static_configs.getLength(); i++){
		DEBUG("Storing static camera config \'%s\' in image header.",static_configs[i].getName());
		switch (static_configs[i].getType()){

			case Setting::TypeInt:
				image->setInfoField(static_configs[i].getName(), (int)static_configs[i]);
				break;
			case Setting::TypeInt64:
				image->setInfoField(static_configs[i].getName(), (long)static_configs[i]);
				break;
			case Setting::TypeFloat:
				image->setInfoField(static_configs[i].getName(), (float)static_configs[i]);
				break;
			case Setting::TypeBoolean:
				image->setInfoField(static_configs[i].getName(), (bool)static_configs[i]);
				break;
			case Setting::TypeString:
				image->setInfoField(static_configs[i].getName(), (const char*)static_configs[i]);
				break;
			default:
			WARN("Unknown data type for static camera config \'%s\' on line %d of config file.",static_configs[i].getName(), static_configs[i].getSourceLine());
			continue;
		}
	}


	//now do the dynamic configs - these need to be set on the camera as well as being stored
	//in the image header.
	Setting &dynamic_configs = camera_cfg["dynamic"];

	for (int i=0; i< dynamic_configs.getLength(); i++){
		DEBUG("Updating dynamic camera config \'%s\' on camera and storing in image header.",dynamic_configs[i].getName());
		switch (dynamic_configs[i].getType()){

			case Setting::TypeInt:
			case Setting::TypeInt64:
				if (setConfig(dynamic_configs[i].getName(),(long)dynamic_configs[i])){
					continue; //probably the wrong data type for this node - so skip it
				}
				image->setInfoField(dynamic_configs[i].getName(), (long)dynamic_configs[i]);
				break;
			case Setting::TypeFloat:
				if (setConfig(dynamic_configs[i].getName(),(double)dynamic_configs[i])){
					continue; //probably the wrong data type for this node - so skip it
				}
				image->setInfoField(dynamic_configs[i].getName(), (float)dynamic_configs[i]);
				break;
			case Setting::TypeBoolean:
				if (setConfig(dynamic_configs[i].getName(),(bool)dynamic_configs[i])){
					continue; //probably the wrong data type for this node - so skip it
				}
				image->setInfoField(dynamic_configs[i].getName(), (bool)dynamic_configs[i]);
				break;
			case Setting::TypeString:
				if (setConfig(dynamic_configs[i].getName(),(const char *)dynamic_configs[i])){
					continue; //probably the wrong data type for this node - so skip it
				}
				image->setInfoField(dynamic_configs[i].getName(), (const char*)dynamic_configs[i]);
				break;
			default:
				WARN("Unknown data type for dynamic camera config \'%s\' on line %d of config file.",dynamic_configs[i].getName(), dynamic_configs[i].getSourceLine());
			continue;
		}
	}

	//record the temperature settings in the image header - the case number will already
	//have been saved, but that is a bit cryptic so store a string representation too
	long cur_case;
	std::stringstream ss;
	getConfig("CurrentCase",cur_case);
	DEBUG("Using case %d on camera.",cur_case);
	float T_l,T_u;
	setConfig("QueryCase", cur_case);
	getConfig("QueryCaseLowLimit",T_l);
	getConfig("QueryCaseHighLimit",T_u);
	ss << T_l << "-" << T_u << "K";
	image->setInfoField("temperature_range", ss.str());

	DEBUG("Temperature range of case = %s",ss.str().c_str());

	return 0;
}


IRImage * SC645Camera::captureImage(void){

	if (this->SimpleCamera::captureImage() != NULL){
		//SimpleCamera::captureImage() updates the data stored in this->image (because it shares a pointer with SimpleCamera::image)

		//occasionally the camera seems to set itself to a different pixel format, which
		//breaks everything - so double check that it is still recording in 16bit and if
		//not then change it back.
		if(image->getPixelFormat() != ARV_PIXEL_FORMAT_MONO_16){
			WARN("Image from camera had unexpected pixel format - setting camera back to 16bit.");
			setConfig("PixelFormat","Mono16");
			return NULL;
		}
		return image;
	} else {
		return NULL;
	}
}


int SC645Camera::autofocus(void){

	ArvGcNode *node = arv_gc_get_node (genicam, "AutoFocus");
	GError *err = NULL;

	if(node == NULL){
		WARN("Failed to get node for AutoFocus");
		return 1;
	}

	arv_gc_command_execute(ARV_GC_COMMAND(node), &err);
	if (err != NULL){
		WARN("Failed to autofocus. %s", err->message);
		g_error_free(err);
		return 1;
	}

	//wait for focusing to finish.
	sleep(2);
	return 0;
}


int SC645Camera::setFocusPos(int focus_pos){
	//max distance focus = 0,
	// min distance focus = 65535
	//if focus_position is -1 then autofocus is used
	DEBUG("Setting the focus position to %d",focus_pos);

	//update the stored focus position - used for the periodic refocusing
	focus_position = focus_pos;

	if(focus_pos == -1){
			if (autofocus()){
				return 1;
			}else{
				return 0;
			}
	}

	if((focus_pos < 0) && (focus_pos > 65535)){
		WARN("Invalid value \'%d\'for focus position, must be in the range 0-65535 or -1 (for autofocus).", focus_pos);
		return 1;
	}
	
	//check if the camera is already at the focus position that we want
	long current_focus_pos;
	if(! getConfig("FocusPos", current_focus_pos)){
		if (current_focus_pos == focus_pos){
			return 0;
		}
	}

	//sometimes the camera defaults to having a focus speed of 0. This is unhelpful, since the focusing
	//then doesn't work properly. Just in case, we set the focus speed here. 50 seems to be a good compromise
	//between speed and accuracy.
	if (setConfig("FocusSpeed", (long) 50)){
		return 1;
	}
	
	//small adjustments to the focus don't work, so we always move to one of the focus extremes first
	//and then back to the requested position
	long focus_extreme;
	if (focus_pos <= 32767){
		focus_extreme = 65535;
	} else {
		focus_extreme = 0;
	}
	
	if (setConfig("FocusPos", focus_extreme)){
		return 1;
	}
	
	sleep(1); //wait for the motor to finish moving
			
	if (setConfig("FocusPos", (long)focus_pos)){
		return 1;
	}

	sleep(2); //wait for the motor to finish moving completely

	//sanity check
#if VERBOSITY==2
	long check;
	getConfig("FocusPos", check);
	if (check != focus_position){
		WARN("Failed to set focus to desired position (%d). It is currently set at %ld.", focus_pos, check);
		return 1;
	}
#endif

	return 0;
}


int SC645Camera::refocus(void){
	return setFocusPos(focus_position);
}


int SC645Camera::performNUC(void){

	ArvGcNode *node = arv_gc_get_node (genicam, "NUCAction");
	GError *err = NULL;

	if(node == NULL){
		WARN("Failed to get node for NUC");
		return 1;
	}

	arv_gc_command_execute(ARV_GC_COMMAND(node), &err);

	if (err != NULL){
		WARN("Failed to perforn NUC. %s", err->message);
		g_error_free(err);
		return 1;
	}

	sleep(1);
	return 0;
}
